/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 */
#include "runtime.h"
#include "event.h"
#include "list.h"
#include "pool.h"
#include "trace.h"


struct event_object {
	/* List of event_handler HANDLEs listening to us */
	LIST event_list;
	/* List of event_handler HANDLEs we are listening for */
	LIST handler_list;

	/* data for us as an event raiser */
	event_filter_func_t filter_func;
	event_filter_check_func_t filter_check_func;

	/* Data and consistancy protection */
	int grabbed; /* Is it being worked on */
	int deleted; /* Is it scheduled for deletion? */
};

struct event_handler {
	POOL p;
	/* The event object that will raise the code */
	HANDLE event_obj;
	/* The handler that will respond */
	HANDLE handler_obj;

	/* The code we are listening for */
	int code;
	/* Filter data */
	void *filter;
	
	void *callback_data;
	event_callback_t callback_func;

	/* statistics */
	int calls_total;
	int calls_this_time;

	/* Data and consistancy protection */
	int grabbed; /* Is it being worked on */
	int deleted; /* Is it scheduled for deletion? */

};


inline static int event_set_errno(int errno);

static int event_errno;
static obj_type_h event_obj_type = obj_type_invalid;

int event_get_errno()
{
	return event_errno;
}
static int event_set_errno(int errno)
{
	event_errno = errno;
	return event_errno;
}


/* event_register_obj(obj,name)
 * Paramaters:
 * obj:
 * 	This is the object that the event object is to be associated
 * 	with.
 * Description:
 * 	This function sets up an object (see object.h) to be able
 * 	to raise and receive events.
 * Returns:
 * 	The error code on failure, or event_err_ok on success.
 * Error Codes:
 * 	event_err_mem: An internal memory problem
 * 	event_err_not_object: The handle passed doesn't refer to an
 * 		existing object.
 * 	event_err_already: The object has already been setup as
 * 		an event object.
 */ 	
int event_register_obj(HANDLE obj)
{
	int r;
	struct event_object *event_obj;

	if (event_obj_type == obj_type_invalid) {
		r = object_register_type(&event_obj_type, "Event Obj");
		if (r != obj_err_ok) {
			trace(T_EVENT,T_ERR,"Failed to register "
					"object type for event obj "
					"err=%d",r);
			return event_err_mem;
		}
	}

	event_obj = object_new(event_obj_type, sizeof(struct event_object));
	if (event_obj == NULL) {
		trace(T_EVENT,T_ERR,"Failed to create new event object "
				"err code = %d",object_get_errno());
		return event_err_mem;
	}

	r = object_attach(obj, event_obj);
	if (r != obj_err_ok) {
		trace(T_EVENT,T_ERR,"Failed to attach event object to "
				"parent object. err code = %d",r);
		/* We try to return a valid, and appropriate error code */
		switch(r) {
		case obj_err_invalid_obj: return event_err_invalid_obj;
		case obj_err_existing: return event_err_already;
		default: return event_err_eieio;
		}
	}

	return event_err_ok;				
}

/* event_deregister_obj(obj)
 * Params:
 * obj:
 *      Specifies the object that is to loose its event object
 *      abilities.
 * Desc:
 * 	This function does the oppisite of event_register_obj()
 * 	It removes the code from the object that allows it to
 * 	raise and receive events. 
 * 	All listeners to this object and all callbacks in other event
 * 	objects from this object are are obviously removed.
 * Returns:
 * 	Returns the error code, or event_err_ok if no error.
 * Error Codes:
 * 	event_err_ok: No error
 * 	event_err_invalid_event_obj: The object passed isn't an 
 * 		event object.
 * 	event_err_mem: Some internal memory error occured.
 */
int event_deregister_obj(HANDLE obj)
{

	/* TODO */
	/* Find the event obj */
	/* Clean the handlers/callbacks */
	/* Detach the event obj from the parent */
	/* destroy the event obj */
}




/* event_add_handler(event_obj, handler_obj,
 * 	event_code, filter, 
 * 	handler_func, handler_data)
 * 
 * Parameters
 * ----------
 * event_obj:	
 * 	The event object to add the handler to.
 *
 * handler_obj:
 * 	The object to receive the event.
 * 	
 * event_code:
 * 	The event code to listen for. (See the discussion on event_codes)
 * 	
 * filter:
 * 	The event depedant filter data. (See the documentation for the
 * 	event to determine what type, if any, the filter should be.
 * 	
 * handler_func:
 * 	The function to call when the event_code is raised by the event_obj.
 * 	This function must be of the type event_callback_t.
 *
 * handler_data:
 * 	Extra data that is passed to the handler_func when the event is
 * 	raised.
 * 	
 * Description
 * -----------
 * 	Adds an event handler to the specified event object. The
 * 	handler is called when the specified event_code is raised by
 * 	the event object. If there is filter data specified, and the
 * 	event object has specified an event filter, then the event
 * 	filter is called first to determine if the specified 
 * 	handler_func should be called.
 * 	The handler_func is passed the event object, the handler object,
 * 	the event code,	the event dependant data (EDD), 
 * 	and the handler_data.
 *
 * Returns
 * -------
 * 	Returns event_err_ok on success, or an error code on failure.
 * 
 * Error Codes
 * -----------
 *  	event_err_not_event_obj: Either the event_obj or the handler_obj
 *  		are not an event object (eg it has not be registered 
 *  		with event_register first).
 *  	event_err_invalid_callback: The specified callback is not
 *  		valid.
 *  	event_err_bad_filter: The filter specified is not in the
 *  		correct format (did not pass the event_check_filter
 *  		function of the event object).
 *  	event_err_invalid_event: The event code specified is not
 *  		a valid event code.
 */

int event_add_handler(HANDLE obj, HANDLE handler_obj,
		int event_code,
		void *filter, /* Event dependant filter */
		event_callback_t handler_func,
		void *handler_data);

/* Delete all events in a specified event object */
int event_del_handler_by_event_obj(HANDLE event_obj);
/* Delete all events in a specified event object that have the specified
 * event code 
 */
int event_del_handler_by_event_code(HANDLE event_obj, int event_code);
/* Delete all callbacks setup by the specified handler_obj
 */
int event_del_handler_by_handler_obj(HANDLE handler_obj); 

/* Delete a specified callback, specified by event_obj, handler_obj,
 * and event_code
 */
int event_del_handler(HANDLE event_obj, HANDLE handler_obj,
		int event_code);



int event_raise(HANDLE event_obj, int event_code, void *event_edd);

/* Sets the event filter for the event object */
int event_set_filter(HANDLE event_obj, event_filter_func_t func);

int event_set_check_filter(HANDLE event_obj, 
		event_filter_check_func func);











/************** old code ***************/





struct eventnode
{
	unsigned int code;
	event_callback_t func;
	void *filter;
	void *appdata;
};

struct eventobj
{
	POOL pool;
	struct list *listeners;
	void *appdata;
	event_filter_func_t filter_func;
};

static int event_compare(struct eventnode *what, struct eventnode *node);


int event_stat_global_raises;


int event_add_listener(struct eventobj *what, unsigned int code,
		void *filter,
		event_callback_t func, void *appdata)
{
	struct eventnode *en;
	ASSERT(what);
	ASSERT(func);

	en = XMALLOC(what->pool,struct eventnode);
	en->func = func;
	en->filter = filter;
	en->code = code;
	en->appdata = appdata;
	list_add_front(what->listeners,en);
	return 0;
}


/* Remove any matching code.
 * If NULL is passed for func or appdata, it is assumed to mean match any.
 */
int event_del_listener_by_data(struct eventobj *obj,
			unsigned int code,
			void *filterdata,
			event_callback_t func,
			void *appdata)
{
	static struct eventnode match,*en;
	int found=0;

	match.code = code;
	match.func = func;
	match.filter = filterdata;
	match.appdata = appdata;
	while(1) {
		en = list_find(obj->listeners, &match, 
				(list_compare_t)event_compare);
		if (en == NULL)
			break;
		list_del_current(obj->listeners);
		found++;
	}
	return found ? found : -1;
}



int event_del_listener(struct eventobj *what, unsigned int code,
		void *filter,
		event_callback_t func)
{
	static struct eventnode match,*en;
	int found;

	match.code = code;
	match.func = func;
	match.filter = filter;
	match.appdata = NULL;
	while(1) {
		en = list_find(what->listeners, &match, 
				(list_compare_t)event_compare);
		if (en == NULL)
			break;
		list_del_current(what->listeners);
		found++;
	}
	return found ? found : -1;

}
int event_compare(struct eventnode *what,struct eventnode *node)
{
	
	if (what->code != event_code_any && what->code != node->code)
		return -1;
	if (what->func != NULL && what->func != node->func)
		return -1;
	if (what->filter != NULL && what->filter != node->filter)
		return -1;
	if (what->appdata != NULL && what->appdata != node->appdata)
		return -1;

	return 0;
}
int event_callback(struct list *l, struct eventnode *node, 
		struct eventmsg *event)
{

	if (node->code == event_code_any ||
			node->code == event->code) {
		event->appdata = node->appdata;
		if (event->obj->filter_func && node->filter) {
			if (event->obj->filter_func(event,node->filter) == -1)
				return 0;
		}
		node->func(event);
		event->count++;
	}
	return 0;
}

int event_set_filter_func(struct eventobj *obj, event_filter_func_t func)
{
	obj->filter_func = func;
	return 0;
}

struct eventobj *event_create_obj(POOL parent, void *appdata)
{
	POOL p;
	struct eventobj *obj;
	p = pool_new(parent);
	obj = XMALLOC(p,struct eventobj);
	obj->pool = p;
	obj->listeners = list_create(p);
	obj->appdata = appdata;
	return obj;
}
	
void event_destroy_obj(struct eventobj *obj)
{
	POOL p;
	ASSERT(obj);
	p = obj->pool;
	ASSERT(p != NULL);
	obj->pool = NULL;
	pool_destroy(p);
}

/* event_raise(obj, code, msgdata)
 * Raises an event within the specified object with the specified code.
 * Event listeners (registered with event_add_listener) that
 * match the specified code, and pass the optional filter (registered
 * with event_set_filter_func) are called with the msgdata as a parameter
 * (see event_add_listener).
 *
 * Returns -1 on error or the number of listeners that were called (0
 * if there were no matches).
 */

int event_raise(struct eventobj *obj, unsigned int code, void *msgdata)
{
	struct eventmsg *msg;
	int r;
	event_stat_global_raises++;
	msg = (struct eventmsg *) malloc(sizeof(struct eventmsg));
	msg->obj = obj;
	msg->msg = msgdata;
	msg->code = code;
	msg->count = 0;
	r = list_foreach(obj->listeners,
			(list_callback_t)event_callback,
			msg);
	if (r != -1)
		r = msg->count;
	free(msg);
	return r;
}


